
#include <linux/kernel.h>
#include <comedi_module.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <linux/malloc.h>

/*

  Computer Boards CIO-DAC 08/AOL A/D adapter

  hack by David Schleef <ds@stm.lbl.gov>

  mostly stolen from Warren J. Jasper <wjasper@tx.ncsu.edu>


 * Copyright (C) 1998  Warren Jasper
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".


*/


#define CIO_SIZE 0x10

#define LSB_AND_CHNLS	0	/* A/D Data & Channel Register      */
#define MSB_DATA_BYTE	1
#define STATUS_REG	2	/* Channel Mux Scan Limits Register */
#define GAIN_REG	3	/* Programmable Gain Register       */
#define COUNTER_0_DATA	4	/* Counter 0 Register               */
#define COUNTER_1_DATA	5	/* Counter 1 Register               */
#define COUNTER_2_DATA	6	/* Counter 2 Register               */
#define COUNTER_CONTROL	7	/* Conter Control  Register         */
#define DA_CHAN0_LSB	8	/* DAC 0 Low Byte                   */
#define DA_CHAN0_MSB	9	/* DAC 0 High Byte                  */
#define DA_CHAN1_LSB	10	/* DAC 1 Low Byte                   */
#define DA_CHAN1_MSB	11	/* DAC 1 High Byte                  */
#define DIO_PORTA	12	/* Port A 8 bit I/O of 8255         */
#define DIO_PORTB	13	/* Port B 8 bit I/O of 8255         */
#define DIO_PORTC	14	/* Port C 4+4 bit of 8255           */
#define DIO_CNTRL_REG	15	/* Mode and Direction Control Reg.  */

/*************************************************************************
* STATUS_REG         base_reg+2             Status Register              *
**************************************************************************/

/* Read */
#define MUX0  0x01      /* Current multiplexor channel */
#define MUX1  0x02      /* Current multiplexor channel */
#define MUX2  0x04      /* Current multiplexor channel */
#define IRQ   0x08      /* 1 = positive edge detected  */
#define IP1   0x10      /* digial input line IP1       */
#define IP2   0x20      /* digial input line IP2       */
#define IP3   0x40      /* digial input line IP3       */
#define EOC   0x80      /* 1=A/D busy, 0=not busy      */

/* Write */
#define INTE  0x08     /* 1=enable interrupts, 0=disable */
#define OP1   0x10     /* digital output line OP1        */
#define OP2   0x20     /* digital output line OP2        */
#define OP3   0x40     /* digital output line OP3        */
#define OP4   0x80     /* digital output line OP4        */

/*************************************************************************
* Constants for dealing with the 8254 counters                           *
**************************************************************************/

#define MODE0 0x0
#define MODE1 0x2
#define MODE2 0x4
#define MODE3 0x6
#define MODE4 0x8
#define MODE5 0xa

#define C0 0x00
#define C1 0x40
#define C2 0x80

#define LATCH    0x00
#define LSBONLY  0x10
#define MSBONLY  0x20
#define LSBFIRST 0x30

#define S0       0x00
#define S1       0x02


static int cio_rem(comedi_device *dev);


typedef struct{
	int dio;
	int aip[16];
}cio_private;
#define devpriv ((cio_private *)dev->private)


static int cio_ai(comedi_device *dev,comedi_trig *it)
{
	comedi_sample s;
	int i;
	
	switch(it->trigsrc){
	case COMEDI_TRIGNOW:
		outb_p(it->chan | devpriv->dio,dev->iobase+STATUS_REG);
		outb_p(1,dev->iobase+MSB_DATA_BYTE);
		for(i=0;i<1000;i++){
			if(!(inb_p(dev->iobase+STATUS_REG) & EOC)){
				s.data=inb(dev->iobase+LSB_AND_CHNLS);
				s.data|=inb(dev->iobase+MSB_DATA_BYTE)<<8;
				s.chan=it->chan;
				it->data=s.data;
				report_sample(dev,&s);
				return 0;
			}
		}
		return -ETIME;
	default:
		return -EINVAL;
	}
}

static int cio_aisp(comedi_device *dev,comedi_param *it)
{
	switch(it->pnum){
	case COMEDI_RANGE:
		if(it->pval<0 || it->pval>=9)return -EINVAL;
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_RANGE,
			0,it->pval);
		devpriv->aip[it->chan]=it->pval;
		return 0;
		break;
	default:
		return -EINVAL;
	}
}

static int cio_ao(comedi_device *dev,comedi_trig *it)
{
	int lsb,msb;
	
	switch(it->trigsrc){
	case COMEDI_TRIGNOW:
		lsb=it->data&0xff;
		msb=(it->data>>8)&0xf;
		if(it->chan==8){
			outb(lsb,dev->iobase+DA_CHAN0_LSB);
			outb(msb,dev->iobase+DA_CHAN0_MSB);
		}else{
			outb(lsb,dev->iobase+DA_CHAN1_LSB);
			outb(msb,dev->iobase+DA_CHAN1_MSB);
		}
		return 0;
	default:
		return -EINVAL;
	}
}

static int cio_do(comedi_device *dev,comedi_trig *it)
{
	switch(it->trigsrc){
	case COMEDI_TRIGNOW:
		devpriv->dio=it->data<<4;
		outb_p(devpriv->dio,dev->iobase+STATUS_REG);
		return 0;
	default:
		return -EINVAL;
	}
}

static int cio_8255(comedi_device *dev,comedi_trig *it)
{
	return -EINVAL;
}

static int cio_8255sp(comedi_device *dev,comedi_param *it)
{
	return -EINVAL;
}

static int cio_di(comedi_device *dev,comedi_trig *it)
{
	switch(it->trigsrc){
	case COMEDI_TRIGNOW:
		it->data=(inb(dev->iobase+STATUS_REG)>>4)&0x7;
		return 0;
	default:
		return -EINVAL;
	}
}

int cio_init(comedi_device *dev,comedi_devconfig *it)
{
	int i,j;
	
	if(strcmp("cio-das8-aol",it->name))
		return 0;
	
	printk("comedi%d: cio: 0x%04x",dev->minor,it->iobase);
	if(check_region(it->iobase,CIO_SIZE)<0){
		printk(" I/O port conflict\n");
		return -EIO;
	}
	dev->iobase=it->iobase;
	dev->iosize=CIO_SIZE;
	dev->irq=0;
	
	dev->chan=8+2+3;
	
	devpriv=kmalloc(sizeof(cio_private)+dev->chan*sizeof(comedi_chinfo),
		GFP_KERNEL);
	if(!devpriv)
		return -ENOMEM;
	
	memset(devpriv,0,sizeof(cio_private)+dev->chan*sizeof(comedi_chinfo));
	request_region(it->iobase,CIO_SIZE,"cio-das8-aol");
	
	dev->chinfo=(void *)(devpriv+1);
	for(i=0;i<8;i++){
		dev->chinfo[i].trig=cio_ai;
		dev->chinfo[i].sp=cio_aisp;
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_BITS,COMEDI_RDONLY,12);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_MAXDATA,COMEDI_RDONLY,0xfff);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_ANALOG);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_INPUT);
		addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE,0,0);
	}
	for(j=0;j<2;j++,i++){
		dev->chinfo[i].trig=cio_ao;
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_BITS,COMEDI_RDONLY,12);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_MAXDATA,COMEDI_RDONLY,0xfff);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_ANALOG);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_OUTPUT);
		addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE,0,0);
	}
	for(j=0;j<3;j++,i++){
		dev->chinfo[i].trig=cio_8255;
		dev->chinfo[i].sp=cio_8255sp;
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_BITS,COMEDI_RDONLY,8);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_DIGITAL);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_BIDIR);
	}
	dev->chinfo[i].trig=cio_di;
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_BITS,COMEDI_RDONLY,3);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_DIGITAL);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_INPUT);
	i++;
	
	dev->chinfo[i].trig=cio_do;
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_BITS,COMEDI_RDONLY,4);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_DIGITAL);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_OUTPUT);
	
	/* zero AO */
	
	outb(0,dev->iobase+DA_CHAN0_LSB);
	outb(0,dev->iobase+DA_CHAN0_MSB);
	outb(0,dev->iobase+DA_CHAN1_LSB);
	outb(0,dev->iobase+DA_CHAN1_MSB);
	
	outb_p(0,dev->iobase+DIO_PORTA);
	outb_p(0,dev->iobase+DIO_PORTB);
	outb_p(0,dev->iobase+DIO_PORTC);
	
	comedi_initbuf(&dev->rdbuf);
	
	dev->rem=cio_rem;
	
	printk("\n");

	return 1;
}

static int cio_rem(comedi_device *dev)
{
	int i;
	
	printk("comedi%d: cio: remove\n",dev->minor);
	
	for(i=0;i<dev->chan;i++)
		delparamlist(dev->chinfo[i].paramlist);
	kfree(devpriv);
	release_region(dev->iobase,dev->iosize);
	
	comedi_freebuf(&dev->rdbuf);
	
	return 0;
}

