/*
    modules/dt282x.c
    hardware driver for Data Translation DT2821 series

    COMEDI - Linux Control and Measurement Device Interface
    Copyright (C) 1997-8 David A. Schleef <ds@stm.lbl.gov>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/


#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/timex.h>
#include <linux/timer.h>
#include <asm/io.h>
#include <comedi_module.h>



#define DT2821_SIZE 0x10

#define DT2821_ADCSR 0
#define DT2821_CHANCSR 2
#define DT2821_ADDAT 4
#define DT2821_DACSR 6
#define DT2821_DADAT 8
#define DT2821_DIODAT 10
#define DT2821_SUPCSR 12
#define DT2821_TMRCTR 14

/*
 * flags
 */

/* ADCSR */

#define DT2821_ADCSR_MASK 0xfff0
#define DT2821_ADCSR_VAL 0x7c00

#define DT2821_ADERR 0x8000
#define DT2821_ADCLK 0x0200
#define DT2821_MUXBUSY 0x0100
#define DT2821_ADDONE 0x0080
#define DT2821_IADDONE 0x0040

/* CHANCSR */

#define DT2821_CHANCSR_MASK 0xfff0
#define DT2821_CHANCSR_VAL 0x70f0

#define DT2821_LLE 0x8000

/* DACSR */

#define DT2821_DACSR_MASK 0xffff
#define DT2821_DACSR_VAL 0x7c9c

#define DT2821_DAERR 0x8000
#define DT2821_YSEL 0x0200
#define DT2821_SSEL 0x0100
#define DT2821_DACRDY 0x0080
#define DT2821_IDARDY 0x0040
#define DT2821_DACLK 0x0020
#define DT2821_HBOE 0x0002
#define DT2821_LBOE 0x0001

/* SUPCSR */

#define DT2821_SUPCSR_MASK 0xfcff
#define DT2821_SUPCSR_VAL 0x0000

#define DT2821_DMAD 0x8000
#define DT2821_ERRINTEN 0x4000
#define DT2821_CLRDMADNE 0x2000
#define DT2821_DDMA 0x1000
#define DT2821_BUFFB 0x0200
#define DT2821_SCDN 0x0100
#define DT2821_DACON 0x0080
#define DT2821_ADCINIT 0x0040
#define DT2821_DACINIT 0x0020
#define DT2821_PRLD 0x0010
#define DT2821_STRIG 0x0008
#define DT2821_XTRIG 0x0004
#define DT2821_XCLK 0x0002
#define DT2821_BDINIT 0x0001

/* TMRCTR */

#define DT2821_TMRCTR_MASK 0xff00
#define DT2821_TMRCTR_VAL 0xf000

typedef struct{
	char *name;
	int boardcode_se;
	int boardcode_di;
}boardname_t;

static boardname_t boardnames[]={
	{"dt2821",0,1},
	{"dt2821-f-16se",0,0},
	{"dt2821-f-8di",1,1},
	{"dt2821-g-16se",0,0},
	{"dt2821-g-8di",1,1},
	{"dt2823",2,2},
	{"dt2824-pgh",3,4},
	{"dt2824-pgl",3,4},
	{"dt2825",0,1},
	{"dt2827",5,5},
	{"dt2828",6,6}
};
static int n_boardnames=11;

typedef struct{
	int adbits;
	int adgains;
	int adchans;
	int sedi;
	int dabits;
	int dachans;
}boardtype_t;

boardtype_t boardtypes[]={
	{ 12, 4, 16, COMEDI_AREF_COM, 12, 2 },	/* 2821, 2825, single ended */
	{ 12, 4, 8, COMEDI_AREF_DIFF, 12, 2 },	/* 2821, 2825, differential */
	{ 16, 1, 4, COMEDI_AREF_DIFF, 16, 2 },	/* 2823 */
	{ 12, 4, 16, COMEDI_AREF_COM, 0, 0 },	/* 2824, single ended */
	{ 12, 4, 8, COMEDI_AREF_DIFF, 0, 0 },	/* 2824, differential */
	{ 16, 1, 4, COMEDI_AREF_DIFF, 12, 2 },	/* 2827 */
	{ 12, 1, 4, COMEDI_AREF_COM, 12, 2 },	/* 2828 */
};

typedef struct{
	int ntrig;
	int curadchan;
	int dacsr;
	int adcsr;
	int ntrig100hz;
	int timer_i;
	int trigvar100hz;
	boardtype_t *board;
}dt282x_private;
#define devpriv ((dt282x_private *)dev->private)
#define boardtype (*devpriv->board)

comedi_devinit dt282x_init;
static int dt282x_rem(comedi_device *dev);

static void dt282x_interrupt(int irq,void *d,struct pt_regs * regs)
{
	comedi_sample pt;
	comedi_device *dev=d;
	
	pt.data=(short)inw(dev->iobase+DT2821_ADDAT);
	pt.chan=devpriv->curadchan;

	report_sample(dev,&pt);

	if(--devpriv->ntrig){
		outw(DT2821_STRIG,dev->iobase+DT2821_SUPCSR);
	}else{
		wake_up_interruptible(&dev->rdwait);
	}
}

static void dt282x_timer(comedi_device *dev)
{
	comedi_sample s;
	
	if(!(--devpriv->timer_i)){
		devpriv->timer_i=devpriv->trigvar100hz;
		outw(DT2821_LLE,dev->iobase+DT2821_CHANCSR);
		outw(devpriv->adcsr,dev->iobase+DT2821_ADCSR);
		outw(0,dev->iobase+DT2821_CHANCSR);
		outw(DT2821_PRLD,dev->iobase+DT2821_SUPCSR);
		while(inw(dev->iobase+DT2821_ADCSR)&DT2821_MUXBUSY)
			udelay(5);
		outw(DT2821_STRIG,dev->iobase+DT2821_SUPCSR);
		while(!(inw(dev->iobase+DT2821_ADCSR)&DT2821_ADDONE))
			udelay(5);
		s.data=(short)inw(dev->iobase+DT2821_ADDAT);
		s.chan=devpriv->curadchan;
		report_sample(dev,&s);
	}
}

static int dt282x_ad(comedi_device *dev,comedi_trig *it)
{
	comedi_sample s;
	int i;

	switch(it->trigsrc){
	case COMEDI_TRIGNOW:
		i=getparamval(dev->chinfo[it->chan].paramlist,COMEDI_RANGE);
		outw(DT2821_LLE,dev->iobase+DT2821_CHANCSR);
		outw(DT2821_ADCLK|(i<<4)|(it->chan),dev->iobase+DT2821_ADCSR);
		outw(0,dev->iobase+DT2821_CHANCSR);
		outw(DT2821_PRLD,dev->iobase+DT2821_SUPCSR);
		while(inw(dev->iobase+DT2821_ADCSR)&DT2821_MUXBUSY)
			udelay(5);
		outw(DT2821_STRIG,dev->iobase+DT2821_SUPCSR);
		while(!(inw(dev->iobase+DT2821_ADCSR)&DT2821_ADDONE))
			udelay(5);
		s.data=it->data=(short)inw(dev->iobase+DT2821_ADDAT);
		s.chan=it->chan;
		report_sample(dev,&s);
		return 0;
	case COMEDI_TRIGCLK:
		devpriv->ntrig=it->n;
		devpriv->curadchan=it->chan;
		i=getparamval(dev->chinfo[it->chan].paramlist,COMEDI_RANGE);
		outw(it->trigvar,dev->iobase+DT2821_TMRCTR);
		outw(DT2821_LLE,dev->iobase+DT2821_CHANCSR);
		outw(DT2821_ADCLK|DT2821_IADDONE|(i<<4)|(it->chan),
			dev->iobase+DT2821_ADCSR);
		outw(0,dev->iobase+DT2821_CHANCSR);
		outw(DT2821_PRLD,dev->iobase+DT2821_SUPCSR);
		while(inw(dev->iobase+DT2821_ADCSR)&DT2821_MUXBUSY)
			udelay(5);
		outw(DT2821_STRIG,dev->iobase+DT2821_SUPCSR);
		return 0;
	case COMEDI_TRIG100HZ:
		devpriv->ntrig100hz=it->n;
		devpriv->curadchan=it->chan;
		i=getparamval(dev->chinfo[it->chan].paramlist,COMEDI_RANGE);
		devpriv->adcsr=(i<<4)|(it->chan);
		devpriv->trigvar100hz=it->trigvar;
		return 0;
	default:
		return -EINVAL;
	}
}

static int dt282x_da(comedi_device *dev,comedi_trig *it)
{
	switch(it->trigsrc){
	case COMEDI_TRIGNOW:
		devpriv->dacsr |= DT2821_SSEL;
		if(it->chan&1)devpriv->dacsr |= DT2821_YSEL;
		else devpriv->dacsr&=~DT2821_YSEL;
		outw(devpriv->dacsr,dev->iobase+DT2821_DACSR);
		outw(it->data,dev->iobase+DT2821_DADAT);
		outw(DT2821_DACON,dev->iobase+DT2821_SUPCSR);
		return 0;
	default:
		return -EINVAL;
	}
}

static int dt282x_dio(comedi_device *dev,comedi_trig *it)
{
	comedi_sample s;

	switch(it->io){
	case COMEDI_IO_INPUT:
		s.data=it->data=inw(dev->iobase+DT2821_DIODAT);
		s.chan=it->chan;
		
		report_sample(dev,&s);
		return 0;
	case COMEDI_IO_OUTPUT:
		outw(it->data,dev->iobase+DT2821_DIODAT);
		return 0;
	default:
		return -EINVAL;
	}
}

static int dt282x_adsp(comedi_device *dev,comedi_param *it)
{
	switch(it->pnum){
	case COMEDI_RANGE:
		if(it->pval<0 || it->pval>=4)return -EINVAL;
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_RANGE,
			0,it->pval);
		return 0;
	default:
		return -EINVAL;
	}
}

static int dt282x_dasp(comedi_device *dev,comedi_param *it)
{
	return -EINVAL;
}

static int dt282x_dsp(comedi_device *dev,comedi_param *it)
{
	switch(it->pnum){
	case COMEDI_IOBITS:
		switch(it->pval&0xff){
		case 0xff&COMEDI_IOBITS_INPUT:
			devpriv->dacsr &= ~DT2821_LBOE;
			break;
		case 0xff&COMEDI_IOBITS_OUTPUT:
			devpriv->dacsr |= DT2821_LBOE;
			break;
		default:
			return -EINVAL;
		}
		switch(it->pval&0xff00){
		case 0xff00&COMEDI_IOBITS_INPUT:
			devpriv->dacsr &= ~DT2821_HBOE;
			break;
		case 0xff00&COMEDI_IOBITS_OUTPUT:
			devpriv->dacsr |= DT2821_HBOE;
			break;
		default:
			return -EINVAL;
		}
		outw(devpriv->dacsr,dev->iobase+DT2821_DACSR);
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_IOBITS,0,
			it->pval);
		return 0;
	default:
		return -EINVAL;
	}
}

int dt282x_init(comedi_device *dev,comedi_devconfig *it)
{
	int i,irqs,irq;
	long flags;
	int board;

	board=-1;
	for(i=0;i<n_boardnames;i++){
		if(!strcmp(boardnames[i].name,it->name)){
			if(it->j_ai_differential)
				board=boardnames[i].boardcode_di;
			else board=boardnames[i].boardcode_se;
			break;
		}
	}
	if(board<0)return 0;
	
	printk("comedi%d: dt282x: 0x%04x ",dev->minor,it->iobase);
	if(check_region(it->iobase,DT2821_SIZE)<0){
		printk("I/O port conflict\n");
		return -EIO;
	}
	request_region(it->iobase,DT2821_SIZE,"dt282x");
	dev->iobase=it->iobase;
	dev->iosize=DT2821_SIZE;
	strcpy(dev->board,it->name);

	outw(DT2821_BDINIT,dev->iobase+DT2821_SUPCSR);
	i=inw(dev->iobase+DT2821_ADCSR);
#ifdef DEBUG
printk("%x,%x,%x,%x,%x ",
		inw(dev->iobase+DT2821_ADCSR),
		inw(dev->iobase+DT2821_CHANCSR),
		inw(dev->iobase+DT2821_DACSR),
		inw(dev->iobase+DT2821_SUPCSR),
		inw(dev->iobase+DT2821_TMRCTR));
#endif

	if(
		((inw(dev->iobase+DT2821_ADCSR)&DT2821_ADCSR_MASK)
			!=DT2821_ADCSR_VAL) ||
		((inw(dev->iobase+DT2821_CHANCSR)&DT2821_CHANCSR_MASK)
			!=DT2821_CHANCSR_VAL) ||
		((inw(dev->iobase+DT2821_DACSR)&DT2821_DACSR_MASK)
			!=DT2821_DACSR_VAL) ||
		((inw(dev->iobase+DT2821_SUPCSR)&DT2821_SUPCSR_MASK)
			!=DT2821_SUPCSR_VAL) ||
		((inw(dev->iobase+DT2821_TMRCTR)&DT2821_TMRCTR_MASK)
			!=DT2821_TMRCTR_VAL)){
		printk("not found\n");
		release_region(it->iobase,DT2821_SIZE);
		return -EIO;
	}

	/* should do board test */
	
	irq=it->irq;
	if(irq<0){
		save_flags(flags);
		sti();
		irqs=probe_irq_on();
	
		/* trigger interrupt */

		udelay(100);

		irq=probe_irq_off(irqs);
		restore_flags(flags);
		if(0 /* error */){
			printk("error probing irq (bad) \n");
		}

	}
	dev->irq=0;
	if(irq>0){
		printk("( irq = %d )\n",irq);
		request_irq(irq,dt282x_interrupt,SA_INTERRUPT,"dt282x",dev);
		dev->irq=irq;
	}else if(irq==0){
		printk("(no irq)\n");
	}else{
		printk("(probe returned multiple irqs--bad)\n");
	}
	
	dev->chan=boardtypes[board].adchans+boardtypes[board].dachans+1;

	dev->private=kmalloc(sizeof(dt282x_private)+
		dev->chan*sizeof(comedi_chinfo),GFP_KERNEL);
	if(!dev->private){
		return -ENOMEM;
	}
	devpriv->board=boardtypes+board;
	dev->chinfo=(void *)(devpriv+1);
	memset(dev->chinfo,0,dev->chan*sizeof(comedi_chinfo));
	for(i=0;i<boardtype.adchans;i++){
		dev->chinfo[i].trig=dt282x_ad;
		dev->chinfo[i].sp=dt282x_adsp;
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_BITS,COMEDI_RDONLY,boardtype.adbits);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_MAXDATA,COMEDI_RDONLY,
			(1<<boardtypes[board].adbits)-1);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_ANALOG);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_INPUT);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_RANGE,0,0);
	}
	for(;i<boardtype.adchans+boardtype.dachans;i++){
		dev->chinfo[i].trig=dt282x_da;
		dev->chinfo[i].sp=dt282x_dasp;
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_BITS,COMEDI_RDONLY,boardtype.dabits);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_MAXDATA,COMEDI_RDONLY,(1<<boardtype.dabits)-1);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_ANALOG);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_OUTPUT);
	}

	dev->chinfo[i].trig=dt282x_dio;
	dev->chinfo[i].sp=dt282x_dsp;
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_BITS,COMEDI_RDONLY,16);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_DIGITAL);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_BIDIR);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IOBITGROUP,COMEDI_RDONLY,8);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IOBITS,COMEDI_RDWR,0xffff&COMEDI_IOBITS_INPUT);
	devpriv->dacsr=0;

	comedi_initbuf(&dev->rdbuf);

	dev->rem=dt282x_rem;

	return 1;
}


static int dt282x_rem(comedi_device *dev)
{
	printk("comedi%d: dt282x: remove\n",dev->minor);
	
	kfree(dev->private);

	if(dev->irq){
		free_irq(dev->irq,dev);
	}
	release_region(dev->iobase,dev->iosize);

	comedi_freebuf(&dev->rdbuf);
	
	return 0;
}


