/*
    module/module.c
    comedi kernel module

    COMEDI - Linux Control and Measurement Device Interface
    Copyright (C) 1997-8 David A. Schleef <ds@stm.lbl.gov>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <linux/module.h>

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/fcntl.h>
#include <linux/timer.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <asm/io.h>
#include <comedi_module.h>

comedi_device *comedi_devices;
static struct timer_list timer;

static void comedi_timer(unsigned long d);

static int comedi_ioctl(struct inode * inode,struct file * file,unsigned int cmd,unsigned long arg)
{
	int retval,i;
	kdev_t minor=MINOR(inode->i_rdev);
	comedi_device *dev=comedi_devices+minor;

	switch(cmd){
	case COMEDI_CONFIG: /* for overall configuation */
	{
		comedi_devconfig db;

#ifdef DEBUG
		printk("entering COMEDI_CONFIG\n");
#endif
		if(!suser())
			return -EPERM;
#if OLD_KERNEL
		retval=verify_area(VERIFY_READ,(void *)arg,
			sizeof(comedi_devconfig));
		if(retval)
			return retval;
		memcpy_fromfs(&db,(void *)arg,sizeof(comedi_devconfig));
#else
		memcpy(&db,(void *)arg,sizeof(comedi_devconfig));
#endif
#if DEBUG
		printk("memcpy ok\n");
#endif
		if(dev->rem){
#if DEBUG
			printk("trying to remove existing device %08x\n",
				dev->rem);
#endif
			retval=dev->rem(dev);
			if(retval<0)return retval;
		}
		memset(dev,0,sizeof(comedi_device));
#if DEBUG
		printk("memset ok\n");
#endif
		dev->minor=minor;
		if(!db.name[0])return 0;
#if DEBUG
		printk("scanning devinits\n");
#endif
		for(i=0;i<comedi_maxinits;i++){
			retval=comedi_devinits[i](dev,&db);
			if(retval<0)
				return retval;
			if(retval)return 0;
		}
#if DEBUG
		printk("scanning devinits failed\n");
#endif
		return -EINVAL;
	}
	case COMEDI_GETINFO: /* returns overall info */
	{
		comedi_info it;
		
		it.iobase=dev->iobase;
		it.iosize=dev->iosize;
		it.irq=dev->irq;
		it.minor=dev->minor;
		it.chan=dev->chan;
		strcpy(it.board,dev->board);
		
#if OLD_KERNEL
		retval=verify_area(VERIFY_WRITE,(void *)arg,sizeof(comedi_info));
		if(retval)
			return retval;
		memcpy_tofs((void *)arg,&it,sizeof(comedi_info));
#else
		memcpy((void *)arg,&it,sizeof(comedi_info));
#endif
		return 0;
	}
	case COMEDI_GETPARAM:
	{
		comedi_param param;
		comedi_iparam *p;
		
#if OLD_KERNEL
		retval=verify_area(VERIFY_READ,(void *)arg,sizeof(comedi_param));
		if(retval)return retval;
		memcpy_fromfs(&param,(void *)arg,sizeof(comedi_param));
#else
		memcpy(&param,(void *)arg,sizeof(comedi_param));
#endif
		p=NULL;
		if(param.chan<dev->chan)p=dev->chinfo[param.chan].paramlist;
		while(p){
			if(p->pnum==param.pnum){
				param.flags=p->flags;
				param.pval=p->pval;
#if OLD_KERNEL
				retval=verify_area(VERIFY_WRITE,(void *)arg,
					sizeof(comedi_param));
				if(retval)return retval;
				memcpy_tofs((void *)arg,&param,
					sizeof(comedi_param));
#else
				memcpy((void *)arg,&param,sizeof(comedi_param));
#endif
				return 0;
			}else{
				p=p->next;
			}
		}
		return -EINVAL;
	}
	case COMEDI_SETPARAM:
	{
		comedi_param param;
		
#if OLD_KERNEL
		retval=verify_area(VERIFY_READ,(void *)arg,sizeof(comedi_param));
		if(retval)return retval;
		memcpy_fromfs(&param,(void *)arg,sizeof(comedi_param));
#else
		memcpy(&param,(void *)arg,sizeof(comedi_param));
#endif
		if(param.chan<dev->chan && dev->chinfo[param.chan].sp)
			return dev->chinfo[param.chan].sp(dev,&param);
		return -EINVAL;
	}
	case COMEDI_GETNPARAM:
	{
		comedi_param param;
		comedi_iparam *p;
		
#if OLD_KERNEL
		retval=verify_area(VERIFY_READ,(void *)arg,sizeof(comedi_param));
		if(retval)return retval;
		memcpy_fromfs(&param,(void *)arg,sizeof(comedi_param));
#else
		memcpy(&param,(void *)arg,sizeof(comedi_param));
#endif
		p=NULL;
		if(param.chan<dev->chan)p=dev->chinfo[param.chan].paramlist;
		if(!p)return -EINVAL;
		while(param.pnum--){
			p=p->next;
			if(!p)return -EINVAL;
		}
		param.pnum=p->pnum;
		param.flags=p->flags;
		param.pval=p->pval;
#if OLD_KERNEL
		retval=verify_area(VERIFY_WRITE,(void *)arg,
			sizeof(comedi_param));
		if(retval)return retval;
		memcpy_tofs((void *)arg,&param,
			sizeof(comedi_param));
#else
		memcpy((void *)arg,&param,sizeof(comedi_param));
#endif
		return 0;
	}
	case COMEDI_TRIG: /* triggers conversions */
	{
		comedi_trig it;
		
#if OLD_KERNEL
		retval=verify_area(VERIFY_READ,(void *)arg,sizeof(comedi_trig));
		if(retval)return retval;
		memcpy_fromfs(&it,(void *)arg,sizeof(comedi_trig));
#else
		memcpy(&it,(void *)arg,sizeof(comedi_trig));
#endif
		if(it.chan>=dev->chan)return -EINVAL;
		if(!dev->chinfo[it.chan].trig) /* shouldn't ever occur */
			return -EINVAL;
		retval=dev->chinfo[it.chan].trig(dev,&it);
		if(retval)return retval;
#if OLD_KERNEL
		retval=verify_area(VERIFY_WRITE,(void *)arg,sizeof(comedi_trig));
		if(retval)return retval;
		memcpy_tofs((void *)arg,&it,sizeof(comedi_trig));
#else
		memcpy((void *)arg,&it,sizeof(comedi_trig));
#endif
		return 0;
	}
#if DSP
	case COMEDI_COPY_TO:
	{
		comedi_copy it;
		void *ptr;

		if(!dev->memcpy_to)
			return -EIO;
#if OLD_KERNEL
		retval=verify_area(VERIFY_READ,(void *)arg,sizeof(comedi_copy));
		if(retval)
			return retval;
		memcpy_fromfs(&it,(void *)arg,sizeof(comedi_copy));
		retval=verify_area(VERIFY_READ,(void *)it.src,it.nbytes);
		if(retval)
			return retval;
#endif
		ptr=kmalloc(it.nbytes,GFP_KERNEL);
		if(!ptr)
			return -ENOMEM;
#if OLD_KERNEL
		memcpy_fromfs(ptr,(void *)it.src,it.nbytes);
#else
		memcpy(ptr,(void *)it.src,it.nbytes);
#endif

		retval=dev->memcpy_to(dev,it.dest,ptr,it.nbytes);
		kfree(ptr);
		return retval;
	}
	case COMEDI_COPY_FROM:
	{
		comedi_copy it;
		void *ptr;

		if(!dev->memcpy_from)
			return -EIO;
#if OLD_KERNEL
		retval=verify_area(VERIFY_READ,(void *)arg,sizeof(comedi_copy));
		if(retval)
			return retval;
		memcpy_fromfs(&it,(void *)arg,sizeof(comedi_copy));
		retval=verify_area(VERIFY_WRITE,(void *)it.src,it.nbytes);
		if(retval)
			return retval;
#else
		memcpy(&it,(void *)arg,sizeof(comedi_copy));
#endif
		ptr=kmalloc(it.nbytes,GFP_KERNEL);
		if(!ptr)
			return -ENOMEM;
		retval=dev->memcpy_from(dev,ptr,it.src,it.nbytes);
		if(!retval){
#if OLD_KERNEL
			memcpy_tofs((void *)it.dest,ptr,it.nbytes);
#else
			memcpy((void *)it.dest,ptr,it.nbytes);
#endif
		}
		kfree(ptr);
		return retval;
	}
	case COMEDI_MEMSET:
	{
		comedi_copy it;

		if(!dev->d_memset)
			return -EIO;
#if OLD_KERNEL
		retval=verify_area(VERIFY_READ,(void *)arg,sizeof(comedi_copy));
		if(retval)
			return retval;
		memcpy_fromfs(&it,(void *)arg,sizeof(comedi_copy));
#else
		memcpy(&it,(void *)arg,sizeof(comedi_copy));
#endif
		return dev->d_memset(dev,it.dest,it.nbytes,it.src);
	}
	case COMEDI_EXEC:
		if(!dev->memexec)
			return -EIO;
		return dev->memexec(dev,arg);
	case COMEDI_RESET:
		if(!dev->reset)
			return -EIO;
		return dev->reset(dev,arg);
#endif
	default:
		return -EIO;
	}
}

static int comedi_write(struct inode *inode,struct file *file,const char *buf,int nbytes)
{
	kdev_t minor=MINOR(inode->i_rdev);
	comedi_device *dev=comedi_devices+minor;
	/*struct wait_queue wait={current,NULL};*/
	int retval;

	if(!dev->write)return -EIO;
	if(!nbytes)return 0;

#if OLD_KERNEL
	retval=verify_area(VERIFY_READ,(void *)buf,nbytes);
	if(retval)
		return retval;
#endif
	return dev->write(dev,buf,nbytes);
}

static int comedi_read(struct inode *inode,struct file * file,char *buf,int nbytes)
{
	kdev_t minor=MINOR(inode->i_rdev);
	comedi_device *dev=comedi_devices+minor;
	int n,count=0,retval=0;
	struct wait_queue wait={current,NULL};

	if(!file->private_data)return -EIO;
	if(!nbytes)return 0;

	add_wait_queue(&dev->rdwait,&wait);
	current->state=TASK_INTERRUPTIBLE;
	while(nbytes>0){
		n=copy_from_buffer((comedi_bufptr *)(file->private_data),buf+count,nbytes);
		if(!n){
			if(file->f_flags&O_NONBLOCK){
				retval=-EAGAIN;
				break;
			}
			if(current->signal & ~current->blocked){
				retval=-ERESTARTSYS;
				break;
			}
			schedule();
			continue;
		}
		count+=n;
		nbytes-=n;
	}
	current->state=TASK_RUNNING;
	remove_wait_queue(&dev->rdwait,&wait);

	return (count ? count : retval);
}

/*
static int comedi_select(struct inode * inode,struct file * file,int sel_type,select_table *wait)
{
	return 0;
}
*/

static int comedi_lseek(struct inode *inode,struct file *file,off_t offset,int origin)
{
	return -ESPIPE;
}

static int comedi_open(struct inode *inode,struct file *file)
{
	kdev_t minor=MINOR(inode->i_rdev);
	comedi_device *dev=comedi_devices+minor;

	if(minor>=COMEDI_DEVICES)return -ENODEV;
	if(!comedi_devices[minor].rem && !suser())
		return -ENODEV;
	MOD_INC_USE_COUNT;
	comedi_devices[minor].use_count++;
	
	if(dev->rdbuf.last)
		file->private_data=(void *)comedi_hookbuf(&(dev->rdbuf));
	else file->private_data=NULL;

	return 0;
}

static void comedi_close(struct inode *inode,struct file *file)
{
	kdev_t minor=MINOR(inode->i_rdev);

	if(file->private_data)
		comedi_unhookbuf((comedi_bufptr *)(file->private_data));
	MOD_DEC_USE_COUNT;
	comedi_devices[minor].use_count--;
}

static struct file_operations comedi_fops={
	comedi_lseek,
	comedi_read,
	comedi_write,
	NULL, /* comedi_readdir */
	NULL, /* comedi_poll, */
	comedi_ioctl,
	NULL, /* comedi_mmap */
	comedi_open,
	comedi_close,
	NULL, /* fsync */
	NULL, /* fasync */
	NULL, /* check media change */
	NULL /* revalidate */
};

int init_module(void)
{
	if(register_chrdev(COMEDI_MAJOR,"comedi",&comedi_fops)){
		printk("comedi: unable to get major %d\n",COMEDI_MAJOR);
		return -EIO;
	}
	comedi_devices=(comedi_device *)kmalloc(sizeof(comedi_device)*COMEDI_DEVICES,GFP_KERNEL);
	if(!comedi_devices)
		return -ENOMEM;
	memset(comedi_devices,0,sizeof(comedi_device)*COMEDI_DEVICES);
	init_param();
	printk("comedi: version " COMEDI_VERSION " David Schleef\n");
	
	init_timer(&timer);
	timer.expires=jiffies+1;
	timer.data=0;
	timer.function=comedi_timer;
	add_timer(&timer);
	
	return 0;
}

void cleanup_module(void)
{
	int i;

	if(MOD_IN_USE)
		printk("comedi: module in use -- remove delayed\n");
	del_timer(&timer);
	for(i=0;i<COMEDI_DEVICES;i++){
		if(comedi_devices[i].rem)
			comedi_devices[i].rem(comedi_devices+i);
	}
	cleanup_param();

	unregister_chrdev(COMEDI_MAJOR,"comedi");
}

void report_sample(comedi_device *dev,comedi_sample *s)
{
	copy_to_buffer(&(dev->rdbuf),s,sizeof(comedi_sample));
	if(dev->chinfo[s->chan].lastdata)
		dev->chinfo[s->chan].lastdata->pval=s->data;
}

static void comedi_timer(unsigned long d)
{
	int i;
	
	for(i=0;i<COMEDI_DEVICES;i++){
		if(comedi_devices[i].timer)
			comedi_devices[i].timer(comedi_devices+i);
	}
	timer.expires++;
	add_timer(&timer);
}

