/*
    module/ni_E.c
    COMEDI - Linux Control and Measurement Device Interface
    Copyright (C) 1997-8 David A. Schleef <ds@stm.lbl.gov>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/* the following defines can be used to detect bad behavior of comedi */
/* change undef to define */
#undef DAVE1	/* autodetect */
#undef DAVE2	/* digital i/o */
#undef DAVE3	/* general configuration */
#undef DAVE4	/* calibration */
#undef DAVE5	/* i_timer stuff */

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <linux/malloc.h>
#include <comedi_module.h>
#include <ni_stc.h>

#define NI_SIZE 0x20

#if 0
static char pnp_magic[]={
	0x00,0x00,0x6a,0xb5,0xda,0xed,0xf6,0xfb,
	0x7d,0xbe,0xdf,0x6f,0x37,0x1b,0x0d,0x86,
	0xc3,0x61,0xb0,0x58,0x2c,0x16,0x8b,0x45,
	0xa2,0xd1,0xe8,0x74,0x3a,0x9d,0xce,0xe7,
	0x73,0x39,0 };

#define PNP_CMD 0x479
#define PNP_ADR 0x279
#define PNP_RD 0x203
#endif

static int ni_modebits1[4]={ 0x3000, 0x2000, 0x1000, 0 };
static int ni_modebits2[4]={ 0xf, 0xf, 0x7, 0x7 };

struct ni_boardtype_struct{
	int boardcode;
	char *name;
	int n_adchan;
	int adbits;
	int alwaysdither;
	int limitedgain;
	int caldactype;
	int eepromlayout;
};

static struct ni_boardtype_struct ni_boardtypes[]={
	{44,"at-mio-16e-1"  ,16,12,0,0,1,1 },
	{25,"at-mio-16e-2"  ,16,12,0,0,1,1 },
	{36,"at-mio-16e-10" ,16,12,0,0,1,2 },
	{37,"at-mio-16de-10",16,12,0,0,1,2 },
	{38,"at-mio-64e-3"  ,64,12,0,0,1,1 },
	{39,"at-mio-16xe-50",16,16,1,1,2,3 }
};
#define N_BOARDTYPES 6

static ni_gainlkup[]={ 1, 2, 4, 7, 9, 10, 12, 15 };

/* Additional windowed registers unique to E series */

#define Configuration_Memory_Clear	82
#define ADC_FIFO_Clear			83

/* i/o port offsets */

/* 8 bit registers */
#define Status				0x01
#define Serial_Command			0x0d
#define Misc_Command			0x0f
#define Port_A				0x19
#define Port_B				0x1b
#define Port_C				0x1d
#define Configuration			0x1f
#define Strobes				0x01
#define Channel_A_Mode			0x03
#define Channel_B_Mode			0x05
#define Channel_C_Mode			0x07
#define AI_AO_Select			0x09
#define G0_G1_Select			0x0b

/* 16 bit registers */
#define Configuration_Memory_Low	0x10
#define Configuration_Memory_High	0x12
#define ADC_FIFO_Data_Register		0x1c
#define AO_Configuration		0x16
#define DAC_FIFO_Data			0x1e
#define DAC0_Direct_Data		0x18
#define DAC1_Direct_Data		0x1a


typedef struct{
	int board;
	int dio;
	int ao0p,ao1p;
	int aip[64];
	int lastchan;
	int last_do;
}ni_private;
#define devpriv ((ni_private *)dev->private)
#define boardtype ni_boardtypes[devpriv->board]

comedi_devinit ni_E_init;
static int ni_dio(comedi_device *dev,comedi_trig *ad);
static int ni_ai(comedi_device *dev,comedi_trig *it);
static int ni_rem(comedi_device *dev);
/*static int pnp_init(int iobase);*/
static int ni_read_eeprom(comedi_device *dev,int addr);
static int ni_getboardtype(comedi_device *dev);
static void ni_calibrate(comedi_device *dev);

static int ni_ai(comedi_device *dev,comedi_trig *it)
{
	comedi_sample s;
	int mode,chan;
	int i;
	
	switch(it->trigsrc){
	case COMEDI_TRIGNOW:
		win_out(1,ADC_FIFO_Clear);
		win_out(1,Configuration_Memory_Clear);

		mode=getparamval(dev->chinfo[it->chan].paramlist,COMEDI_AREF);
		chan=ni_modebits1[mode]|(it->chan&ni_modebits2[mode]);

		outw(chan,dev->iobase+Configuration_Memory_High);
		outw(0x8000|devpriv->aip[it->chan],
			dev->iobase+Configuration_Memory_Low);
		win_out(1,AI_Command_1_Register);
		for(i=0;i<1000;i++){
			if(!(inw(AI_Status_1+dev->iobase)&
			  AI_FIFO_Empty_St)){
				win_out(1,ADC_FIFO_Clear);
				break;
			}
		}
	
		win_out(1,AI_Command_1_Register);
		
		for(i=0;i<1000;i++){
			if(!(inw(AI_Status_1+dev->iobase)&
			  AI_FIFO_Empty_St)){
				s.data=inw(dev->iobase+ADC_FIFO_Data_Register);
				if(!(devpriv->aip[it->chan]&0x100))
					s.data^=0x800;
				s.data&=0xfff;
				s.chan=it->chan;
				it->data=s.data;
				report_sample(dev,&s);
				return 0;
			}
		}
		return -ETIME;
	case COMEDI_TRIGCLK:
		devpriv->lastchan=it->chan;
		
		win_out(1,ADC_FIFO_Clear);
		win_out(1,Configuration_Memory_Clear);

		mode=getparamval(dev->chinfo[it->chan].paramlist,COMEDI_AREF);
		chan=ni_modebits1[mode]|(it->chan&ni_modebits2[mode]);

		outw(chan,dev->iobase+Configuration_Memory_High);
		outw(0x8000|devpriv->aip[it->chan],
			dev->iobase+Configuration_Memory_Low);
		
		win_out(AI_Configuration_Start,Joint_Reset_Register);
		win_out((it->n-1)>>16,AI_SC_Load_A_Registers);
		win_out((it->n-1)&0xffff,AI_SC_Load_A_Registers+1);
		win_out(0x20,AI_Command_1_Register);
		/*win_out(0x60,AI_START_STOP_Select_Register);*/
		win_out(0,AI_SI_Load_A_Registers);
		win_out(1,AI_SI_Load_A_Registers+1);
		win_out(0x200,AI_Command_1_Register);
		win_out((it->trigvar>>16),AI_SI_Load_A_Registers);
		win_out((it->trigvar&0xffff),AI_SI_Load_A_Registers+1);
		win_out(1999,AI_SI2_Load_A_Register);
		win_out(1999,AI_SI2_Load_B_Register);
		win_out(0x0100,AI_Mode_2_Register);
		win_out(0x0800,AI_Command_1_Register);
		win_out(0x0300,AI_Mode_2_Register);
		win_out(AI_Configuration_End,Joint_Reset_Register);

		win_out(1,ADC_FIFO_Clear);
		
		win_out(0x1540,AI_Command_1_Register);
		win_out(1,AI_Command_2_Register);

/*
		for(i=0;i<it->n;i++){
			to=0;
			while(inw(AI_Status_1+dev->iobase)&AI_FIFO_Empty_St){
				if(to++>1000)return -ETIME;
				udelay(100);
			}
			s.data=inw(dev->iobase+ADC_FIFO_Data_Register);
			if(!(devpriv->aip[it->chan]&0x100))
				s.data^=0x800;
			s.data&=0xfff;
			s.chan=it->chan;
			report_sample(dev,&s);
		}
*/
		return 0;
	default:
		return -EINVAL;
	}
}

static int ni_aisp(comedi_device *dev,comedi_param *it)
{
	switch(it->pnum){
	case COMEDI_RANGE:
	{
		int pval=it->pval;
		
		if(pval<0 || pval>=16)return -EINVAL;
		if(boardtype.limitedgain){
			if(pval>=8)return -EINVAL;
			pval=ni_gainlkup[it->pval];
		}
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_RANGE,
			0,it->pval);
		devpriv->aip[it->chan]&=~0x107;
		devpriv->aip[it->chan]|=((pval&0x8)<<5);
		devpriv->aip[it->chan]|=(pval&0x7);
		return 0;
	}
	case COMEDI_AREF:
		if(it->pval<0 || it->pval>=4)return -EINVAL;
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_AREF,
			0,it->pval);
		return 0;
	case COMEDI_DITHER:
		if(boardtype.alwaysdither)return -EINVAL;
		if(it->pval<0 || it->pval>=2)return -EINVAL;
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_DITHER,
			0,it->pval);
		devpriv->aip[it->chan]&=~0x200;
		devpriv->aip[it->chan]|=(it->pval<<9);
		return 0;
	default:
		return -EINVAL;
	}
}

static int ni_ao(comedi_device *dev,comedi_trig *it)
{
	unsigned int i=it->data;
	
	switch(it->trigsrc){
	case COMEDI_TRIGNOW:
		if(i>=0x1000)return -EINVAL;
		i^=0x800;
		if(it->chan==16){
			if(!(devpriv->ao0p&1))i^=0x800;
			outw(i,dev->iobase+DAC0_Direct_Data);
		}else{
			if(!(devpriv->ao1p&1))i^=0x800;
			outw(i,dev->iobase+DAC1_Direct_Data);
		}
		return 0;
	default:
		return -EINVAL;
	}
}

static int ni_aosp(comedi_device *dev,comedi_param *it)
{
	switch(it->pnum){
	case COMEDI_RANGE:
		if(it->pval<0 || it->pval>=2)return -EINVAL;
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_RANGE,
			0,it->pval);
		if(it->chan==16){
			devpriv->ao0p&=~1;
			devpriv->ao0p|=(it->pval?1:0);
			outw(devpriv->ao0p,dev->iobase+AO_Configuration);
		}else{
			devpriv->ao1p&=~1;
			devpriv->ao1p|=(it->pval?1:0);
			outw(devpriv->ao1p,dev->iobase+AO_Configuration);
		}
		return 0;
	case COMEDI_DEGLITCH:
		if(it->pval<0 || it->pval>=2)return -EINVAL;
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_DEGLITCH,
			0,it->pval);
		if(it->chan==16){
			devpriv->ao0p&=~2;
			devpriv->ao0p|=(it->pval?2:0);
			outw(devpriv->ao0p,dev->iobase+AO_Configuration);
		}else{
			devpriv->ao1p&=~2;
			devpriv->ao1p|=(it->pval?2:0);
			outw(devpriv->ao1p,dev->iobase+AO_Configuration);
		}
		return 0;
	default:
		return -EINVAL;
	}
}

static int ni_dio(comedi_device *dev,comedi_trig *it)
{
	comedi_sample s;
	int data,mask;

	switch(it->io){
	case COMEDI_IO_INPUT:
		s.data=it->data=
			win_in(DIO_Input_Register)&0xff;
		s.chan=it->chan;
	
		report_sample(dev,&s);
	
		return 0;
	case COMEDI_IO_OUTPUT:
		if(it->n)mask=it->n; else mask=0xff;
		data=(devpriv->last_do&~mask)|(it->data&mask);
		win_out(data,DIO_Output_Register);
		devpriv->last_do=data;
		return 0;
	default:
		return -EINVAL;
	}
}

static int ni_dsp(comedi_device *dev,comedi_param *it)
{
	switch(it->pnum){
	case COMEDI_IOBITS:
		if(it->pval&~0xff)return -EINVAL;
		devpriv->dio=it->pval;
		win_out(devpriv->dio,DIO_Control_Register);
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_IOBITS,0,
			it->pval);
		return 0;
	default:
		return -EINVAL;
	}
}

static void ni_E_timer(comedi_device *dev)
{
	comedi_sample s;
	
	while(!(inw(AI_Status_1+dev->iobase)&AI_FIFO_Empty_St)){
		s.data=inw(dev->iobase+ADC_FIFO_Data_Register);
		if(!(devpriv->aip[devpriv->lastchan]&0x100))
			s.data^=0x800;
		s.data&=0xfff;
		s.chan=devpriv->lastchan;
		report_sample(dev,&s);
		wake_up_interruptible(&dev->rdwait);
	}
}


int ni_E_init(comedi_device *dev,comedi_devconfig *it)
{
	int i,board;

	if(strcmp("ni_E",it->name))
		return 0;

	printk("comedi%d: ni_E: 0x%04x",dev->minor,it->iobase);
	if(check_region(it->iobase,NI_SIZE)<0){
		printk(" I/O port conflict\n");
		return -EIO;
	}
	dev->iobase=it->iobase;
	dev->iosize=NI_SIZE;
	dev->irq=0;

	board=ni_getboardtype(dev);
	if(board<0)return -EIO;
	
	printk(" %s",ni_boardtypes[board].name);
	strcpy(dev->board,ni_boardtypes[board].name);

	dev->chan=ni_boardtypes[board].n_adchan+3;

	devpriv=kmalloc(sizeof(ni_private)+dev->chan*sizeof(comedi_chinfo),
		GFP_KERNEL);
	if(!devpriv){
		return -ENOMEM;
	}
	memset(devpriv,0,sizeof(ni_private)+dev->chan*sizeof(comedi_chinfo));
	devpriv->board=board;
	request_region(it->iobase,NI_SIZE,"ni_E");
	
	dev->chinfo=(void *)(devpriv+1);
	for(i=0;i<boardtype.n_adchan;i++){
		dev->chinfo[i].trig=ni_ai;
		dev->chinfo[i].sp=ni_aisp;
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_BITS,COMEDI_RDONLY,boardtype.adbits);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_MAXDATA,COMEDI_RDONLY,(1<<boardtype.adbits)-1);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_ANALOG);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_INPUT);
		addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE,0,0);
		addparam(&dev->chinfo[i].paramlist,COMEDI_AREF,COMEDI_RDWR,
			COMEDI_AREF_DIFF);
		if(boardtype.alwaysdither){
			addparam(&dev->chinfo[i].paramlist,COMEDI_DITHER,
				COMEDI_RDONLY,1);
			devpriv->aip[i]=0x0100;
		}else{
			addparam(&dev->chinfo[i].paramlist,COMEDI_DITHER,
				COMEDI_RDWR,0);
			devpriv->aip[i]=0x0000;
		}
	}
	dev->chinfo[i].trig=ni_ao;
	dev->chinfo[i].sp=ni_aosp;
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_BITS,COMEDI_RDONLY,12);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_MAXDATA,COMEDI_RDONLY,(1<<12)-1);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_ANALOG);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_OUTPUT);
	addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE,0,0);
	addparam(&dev->chinfo[i].paramlist,COMEDI_DEGLITCH,0,0);
	i++;
	
	dev->chinfo[i].trig=ni_ao;
	dev->chinfo[i].sp=ni_aosp;
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_BITS,COMEDI_RDONLY,12);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_MAXDATA,COMEDI_RDONLY,0xfff);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_ANALOG);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_OUTPUT);
	addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE,0,0);
	addparam(&dev->chinfo[i].paramlist,COMEDI_DEGLITCH,0,0);
	i++;

	dev->chinfo[i].trig=ni_dio;
	dev->chinfo[i].sp=ni_dsp;
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_BITS,COMEDI_RDONLY,8);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_DIGITAL);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_BIDIR);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IOBITGROUP,COMEDI_RDONLY,1);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IOBITS,COMEDI_RDWR,0xff&COMEDI_IOBITS_INPUT);
	devpriv->dio=0;
#ifndef DAVE2
	win_out(devpriv->dio,DIO_Control_Register);
#endif

#ifndef DAVE3
	/*win_out(0x1b80,Clock_and_FOUT_Register);*/ /* set timebase for ai */
	/*win_out(0x1b80,Clock_and_FOUT_Register);*/ /* set timebase for ao */
	win_out(0x1ba0,Clock_and_FOUT_Register);
	win_out(AI_Configuration_Start,Joint_Reset_Register);
	win_out(0x3f80,Interrupt_A_Ack_Register); /* clear interrupts */
	win_out(0x0000,AI_Command_1_Register); /* don't do anything */
	win_out(0xa4a0,AI_Personal_Register); /* ? */
		/* enable STC output signals */
	win_out(0x032e,AI_Output_Control_Register);
		/* listen to start/stop */
	win_out(0x000d,AI_Mode_1_Register);
	win_out(0,AI_Mode_2_Register); /* ? */
	win_out(0,AI_Mode_3_Register); /* ? */
		/* start1 on AI_START1_Pulse */
	win_out(0x0060,AI_Trigger_Select_Register); /* trigger source */
		/* start on SI_TC */
		/* stop on AI_STOP_IN pin */
	win_out(0x29e0,AI_START_STOP_Select_Register);
	win_out(AI_Configuration_End,Joint_Reset_Register);
#endif
#if 0
	win_out(1,ADC_FIFO_Clear);
	win_out(AI_Reset,Joint_Reset_Register);  /* remove? */
	win_out(1,Configuration_Memory_Clear);  /* remove? */
	win_out(1,ADC_FIFO_Clear);  /* remove? */
	win_out(0x1b00,Clock_and_FOUT_Register); /* set timebase */
	win_out(AI_Configuration_Start,Joint_Reset_Register);
	win_out(0x3f80,Interrupt_A_Ack_Register); /* clear interrupts */
	win_out(0x0000,AI_Command_1_Register); /* don't do anything */
	win_out(0x000c,AI_Mode_1_Register); /* listen to start/stop cmds */
	win_out(AI_Configuration_End,Joint_Reset_Register);  /* remove? */
	win_out(AI_Configuration_Start,Joint_Reset_Register);  /* remove? */
	win_out(0x1b80,Clock_and_FOUT_Register); /* move? */
	win_out(0xa4a0,AI_Personal_Register); /* ? */
	win_out(0x032e,AI_Output_Control_Register); /* output signals */
	win_out(0x032e,AI_Output_Control_Register); /* remove? */
	win_out(AI_Configuration_End,Joint_Reset_Register);  /* remove? */
	win_out(AI_Configuration_Start,Joint_Reset_Register);  /* remove? */
	win_out(0,AI_Mode_2_Register); /* ? */
	win_out(0x032e,AI_Output_Control_Register); /* remove? */
	win_out(AI_Configuration_End,Joint_Reset_Register);  /* remove? */
	win_out(0,AI_Mode_3_Register); /* ? */
	win_out(AI_Configuration_Start,Joint_Reset_Register);  /* remove? */
	win_out(0x000d,AI_Mode_1_Register); /* move? */
	win_out(0x0060,AI_Trigger_Select_Register); /* trigger source */
	win_out(AI_Configuration_End,Joint_Reset_Register);  /* remove? */
	win_out(AI_Configuration_Start,Joint_Reset_Register);  /* remove? */
	win_out(0x0060,AI_START_STOP_Select_Register); /* remove? */
	win_out(AI_Configuration_End,Joint_Reset_Register);  /* remove? */
	win_out(AI_Configuration_Start,Joint_Reset_Register);  /* remove? */
		/* start on AI_START1_Pulse */
		/* stop on AI_STOP_IN pin */
	win_out(0x29e0,AI_START_STOP_Select_Register);
	win_out(AI_Configuration_End,Joint_Reset_Register);
	win_out(1,ADC_FIFO_Clear);  /* remove? */
#endif
	
#ifndef DAVE3
	devpriv->ao0p=0x0000;
	outw(devpriv->ao0p,dev->iobase+AO_Configuration);
	devpriv->ao1p=0x0100;
	outw(devpriv->ao1p,dev->iobase+AO_Configuration);
	win_out(AO_Configuration_Start,Joint_Reset_Register);
	win_out(0x2000,AO_Command_1_Register);
	win_out(0,Interrupt_B_Enable_Register);
	win_out(0x0010,AO_Personal_Register);
	outw(0x3f98,dev->iobase+Interrupt_B_Ack);
	win_out(0x1430,AO_Personal_Register);
	win_out(0,AO_Output_Control_Register);
	win_out(0,AO_Start_Select_Register);
	win_out(0,AO_Command_1_Register);
#endif

#ifndef DAVE4
	ni_calibrate(dev);
#endif
	
	comedi_initbuf(&dev->rdbuf);

#ifndef DAVE5
	dev->timer=ni_E_timer;
#endif
	
	dev->rem=ni_rem;

	printk("\n");

	return 1;
}

/* this is not used, since it doesn't seem to work.  use isapnptools
instead. */
#if 0
static int pnp_init(int iobase)
{
	int i;

	/* initiate board */
	for(i=0;i<34;i++)
		outb(pnp_magic[i],PNP_ADR);

	/* reset pnp registers */
	outb(0x02,PNP_ADR);
	outb(0x07,PNP_CMD);
	
	/* initiate board */
	for(i=0;i<34;i++)
		outb(pnp_magic[i],PNP_ADR);

	/* set pnp isolation state */
	outb(0x03,PNP_ADR);
	outb(0x00,PNP_CMD);
	
	/* set pnp configuration state */
	outb(0x06,PNP_ADR);
	outb(0x01,PNP_CMD);
	
	/* assign pnp base address */
	
	outb(0x60,PNP_ADR);
	outb((iobase>>8)&0xff,PNP_CMD);
	outb(0x61,PNP_ADR);
	outb(iobase&0xff,PNP_CMD);
	
	/* assign pnp data port address */

	outb(0x00,PNP_ADR);
	outb(PNP_RD>>2,PNP_CMD);
	
	/* verify base address */
	
	outb(0x60,PNP_ADR);
	i=inb(PNP_RD)<<8;
	outb(0x61,PNP_ADR);
	i|=inb(PNP_RD);
	
printk("i=0x%x ",i);
	if(i!=iobase)return -EIO;
	
	/* activate board */

	outb(0x03,PNP_ADR);
	outb(0x01,PNP_CMD);

	outb(0x07,PNP_ADR);
	outb(0x00,PNP_CMD);

	outb(0x30,PNP_ADR);
	outb(0x01,PNP_CMD);

	outb(0x07,PNP_ADR);
	outb(0x01,PNP_CMD);

	outb(0x30,PNP_ADR);
	outb(0x01,PNP_CMD);

	outb(0x02,PNP_ADR);
	outb(0x02,PNP_CMD);
	
	return 0;
}
#endif


static int ni_rem(comedi_device *dev)
{
	int i;

	printk("comedi%d: ni_E: remove\n",dev->minor);
	
	for(i=0;i<dev->chan;i++)
		delparamlist(dev->chinfo[i].paramlist);
	kfree(devpriv);
	release_region(dev->iobase,dev->iosize);

	comedi_freebuf(&dev->rdbuf);

	return 0;
}

static int ni_read_eeprom(comedi_device *dev,int addr)
{
	int bit;
	int bitstring;

	bitstring=0x0300|((addr&0x100)<<3)|(addr&0xff);
	outb(0x04,dev->iobase+Serial_Command);
	for(bit=0x8000;bit;bit>>=1){
		outb(0x04|((bit&bitstring)?0x02:0),dev->iobase+Serial_Command);
		outb(0x05|((bit&bitstring)?0x02:0),dev->iobase+Serial_Command);
	}
	bitstring=0;
	for(bit=0x80;bit;bit>>=1){
		outb(0x04,dev->iobase+Serial_Command);
		outb(0x05,dev->iobase+Serial_Command);
		bitstring|=((inb(dev->iobase+Status)&0x01)?bit:0);
	}
	outb(0x00,dev->iobase+Serial_Command);
	
	return bitstring;
}

static void ni_write_caldac(comedi_device *dev,int addr,int val)
{
	int loadbit,bits,bit,bitstring;
	
	switch(boardtype.caldactype){
	case 1:
		bits=12;
		bitstring=((addr&0xf)<<8)|(val&0xff);
		loadbit=0x08;
		break;
	case 2:
		if(addr==8){
			bits=12;
			bitstring=val&0xfff;
			loadbit=0x10;
		}else{
			bits=11;
			bitstring=((addr&0x7)<<8)|(val&0xff);
			loadbit=0x08;
		}
		break;
	default:
		return;
	}

	for(bit=1<<(bits-1);bit;bit>>=1){
		outb(((bit&bitstring)?0x02:0),dev->iobase+Serial_Command);
		outb(1|((bit&bitstring)?0x02:0),dev->iobase+Serial_Command);
	}
	outb(loadbit,dev->iobase+Serial_Command);
	outb(0,dev->iobase+Serial_Command);
}

static int ni_getboardtype(comedi_device *dev)
{
#ifndef DAVE1
	int boardcode=ni_read_eeprom(dev,511);
	int i;
	
	for(i=0;i<N_BOARDTYPES;i++){
		if(ni_boardtypes[i].boardcode==boardcode){
			return i;
		}
	}
	printk(" unknown board code %d -- contact author\n",boardcode);
	return -1;
#else
	return 4;
#endif
}

static void ni_calibrate(comedi_device *dev)
{
	switch(boardtype.eepromlayout){
	case 1:
		ni_write_caldac(dev,4,ni_read_eeprom(dev,424));
		ni_write_caldac(dev,1,ni_read_eeprom(dev,423));
		ni_write_caldac(dev,3,ni_read_eeprom(dev,422));
		ni_write_caldac(dev,2,ni_read_eeprom(dev,421));
		break;
	case 2:
		ni_write_caldac(dev,4,ni_read_eeprom(dev,422));
		ni_write_caldac(dev,11,ni_read_eeprom(dev,421));
		ni_write_caldac(dev,1,ni_read_eeprom(dev,420));
		ni_write_caldac(dev,3,ni_read_eeprom(dev,419));
		ni_write_caldac(dev,2,ni_read_eeprom(dev,418));
		break;
	default:
	}
}

