/*
    module/rti800.c
    hardware driver for Analog Devices RTI-800/815 board

    COMEDI - Linux Control and Measurement Device Interface
    Copyright (C) 1998 David A. Schleef <ds@stm.lbl.gov>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/timex.h>
#include <linux/timer.h>
#include <asm/io.h>
#include <comedi_module.h>



#define RTI800_SIZE 16

#define RTI800_CSR 0
#define RTI800_MUXGAIN 1
#define RTI800_CONVERT 2
#define RTI800_ADCLO 3
#define RTI800_ADCHI 4
#define RTI800_DAC0LO 5
#define RTI800_DAC0HI 6
#define RTI800_DAC1LO 7
#define RTI800_DAC1HI 8
#define RTI800_CLRFLAGS 9
#define RTI800_DI 10
#define RTI800_DO 11
#define RTI800_9513A_DATA 12
#define RTI800_9513A_CNTRL 13
#define RTI800_9513A_STATUS 13


/*
 * flags
 */

#define RTI800_BUSY 0x80
#define RTI800_DONE 0x40
#define RTI800_OVERRUN 0x20
#define RTI800_TCR 0x10
#define RTI800_DMA_ENAB 0x08
#define RTI800_INTR_TC 0x04
#define RTI800_INTR_EC 0x02
#define RTI800_INTR_OVRN 0x01


comedi_devinit rti800_init;
static int rti800_rem(comedi_device *dev);
static void rti800_interrupt(int irq,void *dev,struct pt_regs * regs);

typedef struct{
	int junk;
}rti800_private;
#define devpriv ((rti800_private *)dev->private)


#define RTI800_TIMEOUT 100

static int rti800_softtrig(comedi_device *dev,int chan)
{
	comedi_sample s;
	int i,hi,lo,gain;
	unsigned long flags;

	gain=getparamval(dev->chinfo[chan].paramlist,COMEDI_RANGE);
	outb(chan|(gain<<5),dev->iobase+RTI800_MUXGAIN);
	save_flags(flags);
	cli();
	outb(0,dev->iobase+RTI800_CONVERT);
	for(i=0;i<RTI800_TIMEOUT;i++){
		if(inb(dev->iobase+RTI800_CSR)&RTI800_DONE)
			break;
	}
	lo=inb(dev->iobase+RTI800_ADCLO);
	hi=inb(dev->iobase+RTI800_ADCHI);
	restore_flags(flags);

	s.chan=chan;
	s.data=(hi<<8)|lo;
	report_sample(dev,&s);
	
	return s.data;
}

static int rti800_ai(comedi_device *dev,comedi_trig *it)
{
	/*if(devpriv->busy)return -EBUSY;*/

	switch(it->trigsrc){
	case COMEDI_TRIGNOW:
		it->data=rti800_softtrig(dev,it->chan);
		break;
/*
	case COMEDI_TRIGCLK:
		devpriv->ntrig=it->n;
		outb(it->chan|DT2814_ENB|(it->trigvar<<5),
			dev->iobase+DT2814_CSR);
		break;
*/
	default:
		return -EINVAL;
	}
	
	return 0;
} 

static int rti800_ao(comedi_device *dev,comedi_trig *it)
{
	switch(it->chan){
	case 16:
		outb(it->data&0xff,dev->iobase+RTI800_DAC0LO);
		outb(it->data>>8,dev->iobase+RTI800_DAC0HI);
		return 0;
	case 17:
		outb(it->data&0xff,dev->iobase+RTI800_DAC1LO);
		outb(it->data>>8,dev->iobase+RTI800_DAC1HI);
		return 0;
	}
	return -EINVAL;
}

static int rti800_di(comedi_device *dev,comedi_trig *it)
{
	it->data=inb(dev->iobase+RTI800_DI);
	
	return 0;
}

static int rti800_do(comedi_device *dev,comedi_trig *it)
{
	outb(it->data^0xff,dev->iobase+RTI800_DO);
	
	return 0;
}

static int rti800_aisp(comedi_device *dev,comedi_param *it)
{
	switch(it->pnum){
	case COMEDI_RANGE:
		if(it->pval<0 || it->pval>3)return -EINVAL;
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_RANGE,
			0,it->pval);
		return 0;
	default:
		return -EINVAL;
	}
}

int rti800_init(comedi_device *dev,comedi_devconfig *it)
{
	int i,irqs,irq;
	long flags;
	int board;

	board=-1;
	if(!strcmp("rti800",it->name))board=0;
	if(!strcmp("rti815",it->name))board=1;
	if(board<0)return 0;

	printk("comedi%d: rti800: 0x%04x ",dev->minor,it->iobase);
	if(check_region(it->iobase,RTI800_SIZE)<0){
		printk("I/O port conflict\n");
		return -EIO;
	}
	request_region(it->iobase,RTI800_SIZE,"rti800");
	dev->iobase=it->iobase;
	dev->iosize=RTI800_SIZE;

	if(board==0)strcpy(dev->board,"rti800");
	else strcpy(dev->board,"rti815");

/*
	irq=it->irq;
	if(irq<0){
		save_flags(flags);
		sti();
		irqs=probe_irq_on();
	
		outb(0,dev->iobase+DT2814_CSR);

		udelay(100);

		irq=probe_irq_off(irqs);
		restore_flags(flags);
		if(inb(dev->iobase+DT2814_CSR)&DT2814_ERR){
			printk("error probing irq (bad) \n");
		}

		i=inb(dev->iobase+DT2814_DATA);
		i=inb(dev->iobase+DT2814_DATA);
	}
	dev->irq=0;
	if(irq>0){
		printk("( irq = %d )\n",irq);
		request_irq(irq,dt2814_interrupt,0*SA_INTERRUPT,"dt2814",dev);
		dev->irq=irq;
	}else if(irq==0){
		printk("(no irq)\n");
	}else{
		printk("(probe returned multiple irqs--bad)\n");
	}
*/
	
	if(board==0)dev->chan=18;
	else dev->chan=20;

	dev->private=kmalloc(sizeof(rti800_private)+
		dev->chan*sizeof(comedi_chinfo),GFP_KERNEL);
	if(!dev->private){
		return -ENOMEM;
	}
	memset(dev->private,0,sizeof(rti800_private)+
		dev->chan*sizeof(comedi_chinfo));
	dev->chinfo=dev->private+sizeof(rti800_private);
	for(i=0;i<16;i++){
		dev->chinfo[i].trig=rti800_ai;
		dev->chinfo[i].sp=rti800_aisp;
		addparam(&dev->chinfo[i].paramlist,COMEDI_BITS,COMEDI_RDONLY,
			12);
		addparam(&dev->chinfo[i].paramlist,COMEDI_MAXDATA,COMEDI_RDONLY,
			0xfff);
		addparam(&dev->chinfo[i].paramlist,COMEDI_AD,COMEDI_RDONLY,
			COMEDI_AD_ANALOG);
		addparam(&dev->chinfo[i].paramlist,COMEDI_IO,COMEDI_RDONLY,
			COMEDI_IO_INPUT);
		addparam(&dev->chinfo[i].paramlist,COMEDI_LASTDATA,
			COMEDI_RDONLY,0);
		addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE,COMEDI_RDWR,
			0);
		dev->chinfo[i].lastdata=dev->chinfo[i].paramlist;
	}
	
	if(board==1){
		dev->chinfo[i].trig=rti800_ao;
		dev->chinfo[i].sp=NULL;
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_BITS,COMEDI_RDONLY,12);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_MAXDATA,COMEDI_RDONLY,0xfff);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_ANALOG);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_OUTPUT);
		i++;
	
		dev->chinfo[i].trig=rti800_ao;
		dev->chinfo[i].sp=NULL;
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_BITS,COMEDI_RDONLY,12);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_MAXDATA,COMEDI_RDONLY,0xfff);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_ANALOG);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_OUTPUT);
		i++;
	}

	dev->chinfo[i].trig=rti800_di;
	dev->chinfo[i].sp=NULL;
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_BITS,COMEDI_RDONLY,8);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_DIGITAL);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_INPUT);
	i++;

	dev->chinfo[i].trig=rti800_do;
	dev->chinfo[i].sp=NULL;
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_BITS,COMEDI_RDONLY,8);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_DIGITAL);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_OUTPUT);
	i++;

	comedi_initbuf(&dev->rdbuf);

	dev->rem=rti800_rem;
	
	printk("\n");

	return 1;
}


static int rti800_rem(comedi_device *dev)
{
	int i;
	
	printk("comedi%d: rti800: remove\n",dev->minor);
	
	for(i=0;i<dev->chan;i++)
		delparamlist(dev->chinfo[i].paramlist);
	kfree(dev->private);

	if(dev->irq){
		free_irq(dev->irq,dev);
	}
	release_region(dev->iobase,dev->iosize);

	comedi_freebuf(&dev->rdbuf);

	return 0;
}



